@@ -38,6 +38,33 @@
 #include "loopdev.h"
 #include "canonicalize.h"
 
+#define CONFIG_LOOPDEV_DEBUG
+
+#ifdef CONFIG_LOOPDEV_DEBUG
+# include <stdarg.h>
+
+# define DBG(l,x)	do { \
+				if ((l)->debug) {\
+					fprintf(stderr, "loopdev:  [%p]: ", (l)); \
+					x; \
+				} \
+			} while(0)
+
+static inline void __attribute__ ((__format__ (__printf__, 1, 2)))
+loopdev_debug(const char *mesg, ...)
+{
+	va_list ap;
+	va_start(ap, mesg);
+	vfprintf(stderr, mesg, ap);
+	va_end(ap);
+	fputc('\n', stderr);
+}
+
+#else /* !CONFIG_LOOPDEV_DEBUG */
+# define DBG(m,x) do { ; } while(0)
+#endif
+
+
 #define loopcxt_ioctl_enabled(_lc)	(!((_lc)->flags & LOOPDEV_FL_NOIOCTL))
 
 /*
@@ -60,6 +87,7 @@ int loopcxt_set_device(struct loopdev_cxt *lc, const char *device)
 	lc->fd = -1;
 	lc->mode = 0;
 	lc->has_info = 0;
+	lc->info_failed = 0;
 	*lc->device = '\0';
 	memset(&lc->info, 0, sizeof(lc->info));
 
@@ -81,12 +109,18 @@ int loopcxt_set_device(struct loopdev_cxt *lc, const char *device)
 			strncpy(lc->device, device, sizeof(lc->device));
 			lc->device[sizeof(lc->device) - 1] = '\0';
 		}
+		DBG(lc, loopdev_debug("%s successfully assigned", device));
 	}
 
 	sysfs_deinit(&lc->sysfs);
 	return 0;
 }
 
+int loopcxt_has_device(struct loopdev_cxt *lc)
+{
+	return lc && *lc->device;
+}
+
 /*
  * @lc: context
  * @flags: LOOPDEV_FL_* flags
@@ -109,6 +143,8 @@ int loopcxt_set_device(struct loopdev_cxt *lc, const char *device)
  */
 int loopcxt_init(struct loopdev_cxt *lc, int flags)
 {
+	struct stat st;
+
 	if (!lc)
 		return -EINVAL;
 
@@ -116,13 +152,16 @@ int loopcxt_init(struct loopdev_cxt *lc, int flags)
 	lc->flags = flags;
 	loopcxt_set_device(lc, NULL);
 
-	if (!(lc->flags && LOOPDEV_FL_NOSYSFS) &&
+	if (!(lc->flags & LOOPDEV_FL_NOSYSFS) &&
 	    get_linux_version() >= KERNEL_VERSION(2,6,37))
 		/*
 		 * Use only sysfs for basic information about loop devices
 		 */
 		lc->flags |= LOOPDEV_FL_NOIOCTL;
 
+	if (!(lc->flags & LOOPDEV_FL_CONTROL) && !stat(_PATH_DEV_LOOPCTL, &st))
+		lc->flags |= LOOPDEV_FL_CONTROL;
+
 	return 0;
 }
 
@@ -136,6 +175,8 @@ void loopcxt_deinit(struct loopdev_cxt *lc)
 	if (!lc)
 		return;
 
+	DBG(lc, loopdev_debug("de-initialize"));
+
 	free(lc->filename);
 	lc->filename = NULL;
 
@@ -143,6 +184,18 @@ void loopcxt_deinit(struct loopdev_cxt *lc)
 	loopcxt_deinit_iterator(lc);
 }
 
+/*
+ * @lc: context
+ * @enable: TRUE/FALSE
+ *
+ * Enabled/disables debug messages
+ */
+void loopcxt_enable_debug(struct loopdev_cxt *lc, int enable)
+{
+	if (lc)
+		lc->debug = enable ? 1 : 0;
+}
+
 /*
  * @lc: context
  *
@@ -177,12 +230,16 @@ struct sysfs_cxt *loopcxt_get_sysfs(struct loopdev_cxt *lc)
 
 	if (!lc->sysfs.devno) {
 		dev_t devno = sysfs_devname_to_devno(lc->device, NULL);
-		if (!devno)
+		if (!devno) {
+			DBG(lc, loopdev_debug("sysfs: failed devname to devno"));
 			return NULL;
-
-		if (sysfs_init(&lc->sysfs, devno, NULL))
+		}
+		if (sysfs_init(&lc->sysfs, devno, NULL)) {
+			DBG(lc, loopdev_debug("sysfs: init failed"));
 			return NULL;
+		}
 	}
+
 	return &lc->sysfs;
 }
 
@@ -201,6 +258,7 @@ int loopcxt_get_fd(struct loopdev_cxt *lc)
 	if (lc->fd < 0) {
 		lc->mode = lc->flags & LOOPDEV_FL_RDWR ? O_RDWR : O_RDONLY;
 		lc->fd = open(lc->device, lc->mode);
+		DBG(lc, loopdev_debug("open %s", lc->fd < 0 ? "failed" : "ok"));
 	}
 	return lc->fd;
 }
@@ -231,6 +289,8 @@ int loopcxt_init_iterator(struct loopdev_cxt *lc, int flags)
 	if (!lc)
 		return -EINVAL;
 
+	DBG(lc, loopdev_debug("iter: initialize"));
+
 	iter = &lc->iter;
 
 	/* always zeroize
@@ -244,7 +304,7 @@ int loopcxt_init_iterator(struct loopdev_cxt *lc, int flags)
 		/*
 		 * Check for /dev/loop/<N> subdirectory
 		 */
-		if (!(lc->flags && LOOPDEV_FL_DEVSUBDIR) &&
+		if (!(lc->flags & LOOPDEV_FL_DEVSUBDIR) &&
 		    stat(_PATH_DEV_LOOP, &st) == 0 && S_ISDIR(st.st_mode))
 			lc->flags |= LOOPDEV_FL_DEVSUBDIR;
 
@@ -260,11 +320,13 @@ int loopcxt_init_iterator(struct loopdev_cxt *lc, int flags)
  */
 int loopcxt_deinit_iterator(struct loopdev_cxt *lc)
 {
-	struct loopdev_iter *iter = &lc->iter;
+	struct loopdev_iter *iter;
 
 	if (!lc)
 		return -EINVAL;
 
+	DBG(lc, loopdev_debug("iter: de-initialize"));
+
 	iter = &lc->iter;
 
 	free(iter->minors);
@@ -303,6 +365,7 @@ static int loopiter_set_device(struct loopdev_cxt *lc, const char *device)
 	if ((lc->iter.flags & LOOPITER_FL_FREE) && !used)
 		return 0;
 
+	DBG(lc, loopdev_debug("iter: setting device"));
 	loopcxt_set_device(lc, NULL);
 	return 1;
 }
@@ -329,7 +392,6 @@ static int loop_scandir(const char *dirname, int **ary, int hasprefix)
 	dir = opendir(dirname);
 	if (!dir)
 		return 0;
-
 	free(*ary);
 	*ary = NULL;
 
@@ -365,13 +427,15 @@ static int loop_scandir(const char *dirname, int **ary, int hasprefix)
 			tmp = realloc(*ary, arylen * sizeof(int));
 			if (!tmp) {
 				free(*ary);
+				closedir(dir);
 				return -1;
 			}
 			*ary = tmp;
 		}
-		(*ary)[count++] = n;
+		if (*ary)
+			(*ary)[count++] = n;
 	}
-	if (count)
+	if (count && *ary)
 		qsort(*ary, count, sizeof(int), cmpnum);
 
 	closedir(dir);
@@ -391,6 +455,9 @@ int loopcxt_next(struct loopdev_cxt *lc)
 
 	if (!lc)
 		return -EINVAL;
+
+	DBG(lc, loopdev_debug("iter: next"));
+
 	iter = &lc->iter;
 	if (iter->done)
 		return 1;
@@ -477,7 +544,7 @@ struct loop_info64 *loopcxt_get_info(struct loopdev_cxt *lc)
 {
 	int fd;
 
-	if (!lc)
+	if (!lc || lc->info_failed)
 		return NULL;
 	if (lc->has_info)
 		return &lc->info;
@@ -488,7 +555,12 @@ struct loop_info64 *loopcxt_get_info(struct loopdev_cxt *lc)
 
 	if (ioctl(fd, LOOP_GET_STATUS64, &lc->info) == 0) {
 		lc->has_info = 1;
+		lc->info_failed = 0;
+		DBG(lc, loopdev_debug("reading loop_info64 OK"));
 		return &lc->info;
+	} else {
+		lc->info_failed = 1;
+		DBG(lc, loopdev_debug("reading loop_info64 FAILED"));
 	}
 
 	return NULL;
@@ -521,6 +593,8 @@ char *loopcxt_get_backing_file(struct loopdev_cxt *lc)
 			res = strdup((char *) lo->lo_file_name);
 		}
 	}
+
+	DBG(lc, loopdev_debug("get_backing_file [%s]", res));
 	return res;
 }
 
@@ -543,10 +617,11 @@ int loopcxt_get_offset(struct loopdev_cxt *lc, uint64_t *offset)
 		if (lo) {
 			if (offset)
 				*offset = lo->lo_offset;
-			return 0;
+			rc = 0;
 		}
 	}
 
+	DBG(lc, loopdev_debug("get_offset [rc=%d]", rc));
 	return rc;
 }
 
@@ -569,17 +644,153 @@ int loopcxt_get_sizelimit(struct loopdev_cxt *lc, uint64_t *size)
 		if (lo) {
 			if (size)
 				*size = lo->lo_sizelimit;
-			return 0;
+			rc = 0;
 		}
 	}
 
+	DBG(lc, loopdev_debug("get_sizelimit [rc=%d]", rc));
 	return rc;
 }
 
 /*
  * @lc: context
+ * @devno: returns encryption type
  *
- * Returns: 1 of the autoclear flags is set.
+ * Cryptoloop is DEPRECATED!
+ *
+ * Returns: <0 on error, 0 on success
+ */
+int loopcxt_get_encrypt_type(struct loopdev_cxt *lc, uint32_t *type)
+{
+	struct loop_info64 *lo = loopcxt_get_info(lc);
+	int rc = -EINVAL;
+
+	if (lo) {
+		if (type)
+			*type = lo->lo_encrypt_type;
+		rc = 0;
+	}
+	DBG(lc, loopdev_debug("get_encrypt_type [rc=%d]", rc));
+	return rc;
+}
+
+/*
+ * @lc: context
+ * @devno: returns crypt name
+ *
+ * Cryptoloop is DEPRECATED!
+ *
+ * Returns: <0 on error, 0 on success
+ */
+const char *loopcxt_get_crypt_name(struct loopdev_cxt *lc)
+{
+	struct loop_info64 *lo = loopcxt_get_info(lc);
+
+	if (lo)
+		return (char *) lo->lo_crypt_name;
+
+	DBG(lc, loopdev_debug("get_crypt_name failed"));
+	return NULL;
+}
+
+/*
+ * @lc: context
+ * @devno: returns backing file devno
+ *
+ * Returns: <0 on error, 0 on success
+ */
+int loopcxt_get_backing_devno(struct loopdev_cxt *lc, dev_t *devno)
+{
+	struct loop_info64 *lo = loopcxt_get_info(lc);
+	int rc = -EINVAL;
+
+	if (lo) {
+		if (devno)
+			*devno = lo->lo_device;
+		rc = 0;
+	}
+	DBG(lc, loopdev_debug("get_backing_devno [rc=%d]", rc));
+	return rc;
+}
+
+/*
+ * @lc: context
+ * @ino: returns backing file inode
+ *
+ * Returns: <0 on error, 0 on success
+ */
+int loopcxt_get_backing_inode(struct loopdev_cxt *lc, ino_t *ino)
+{
+	struct loop_info64 *lo = loopcxt_get_info(lc);
+	int rc = -EINVAL;
+
+	if (lo) {
+		if (ino)
+			*ino = lo->lo_inode;
+		rc = 0;
+	}
+	DBG(lc, loopdev_debug("get_backing_inode [rc=%d]", rc));
+	return rc;
+}
+
+/*
+ * Check if the kernel supports partitioned loop devices.
+ *
+ * Notes:
+ *   - kernels < 3.2 support partitioned loop devices and PT scanning
+ *     only if max_part= module paremeter is non-zero
+ *
+ *   - kernels >= 3.2 always support partitioned loop devices
+ *
+ *   - kernels >= 3.2 always support BLKPG_{ADD,DEL}_PARTITION ioctls
+ *
+ *   - kernels >= 3.2 enable PT scanner only if max_part= is non-zero or if the
+ *     LO_FLAGS_PARTSCAN flag is set for the device. The PT scanner is disabled
+ *     by default.
+ *
+ *  See kernel commit e03c8dd14915fabc101aa495828d58598dc5af98.
+ */
+int loopmod_supports_partscan(void)
+{
+	int rc, ret = 0;
+	FILE *f;
+
+	if (get_linux_version() >= KERNEL_VERSION(3,2,0))
+		return 1;
+
+	f = fopen("/sys/module/loop/parameters/max_part", "r");
+	if (!f)
+		return 0;
+	rc = fscanf(f, "%d", &ret);
+	fclose(f);
+	return rc == 1 ? ret : 0;
+}
+
+/*
+ * @lc: context
+ *
+ * Returns: 1 if the partscan flags is set *or* (for old kernels) partitions
+ * scannig is enabled for all loop devices.
+ */
+int loopcxt_is_partscan(struct loopdev_cxt *lc)
+{
+	struct sysfs_cxt *sysfs = loopcxt_get_sysfs(lc);
+
+	if (sysfs) {
+		/* kernel >= 3.2 */
+		int fl;
+		if (sysfs_read_int(sysfs, "loop/partscan", &fl) == 0)
+			return fl;
+	}
+
+	/* old kernels (including kernels without loopN/loop/<flags> directory */
+	return loopmod_supports_partscan();
+}
+
+/*
+ * @lc: context
+ *
+ * Returns: 1 if the autoclear flags is set.
  */
 int loopcxt_is_autoclear(struct loopdev_cxt *lc)
 {
@@ -602,7 +813,7 @@ int loopcxt_is_autoclear(struct loopdev_cxt *lc)
 /*
  * @lc: context
  *
- * Returns: 1 of the readonly flags is set.
+ * Returns: 1 if the readonly flags is set.
  */
 int loopcxt_is_readonly(struct loopdev_cxt *lc)
 {
@@ -622,6 +833,67 @@ int loopcxt_is_readonly(struct loopdev_cxt *lc)
 	return 0;
 }
 
+/*
+ * @lc: context
+ * @st: backing file stat or NULL
+ * @backing_file: filename
+ * @offset: offset
+ * @flags: LOOPDEV_FL_OFFSET if @offset should not be ignored
+ *
+ * Returns 1 if the current @lc loopdev is associated with the given backing
+ * file. Note that the preferred way is to use devno and inode number rather
+ * than filename. The @backing_file filename is poor solution usable in case
+ * that you don't have rights to call stat().
+ *
+ * Don't forget that old kernels provide very restricted (in size) backing
+ * filename by LOOP_GET_STAT64 ioctl only.
+ */
+int loopcxt_is_used(struct loopdev_cxt *lc,
+		    struct stat *st,
+		    const char *backing_file,
+		    uint64_t offset,
+		    int flags)
+{
+	ino_t ino;
+	dev_t dev;
+
+	if (!lc)
+		return 0;
+
+	DBG(lc, loopdev_debug("checking %s vs. %s",
+				loopcxt_get_device(lc),
+				backing_file));
+
+	if (st && loopcxt_get_backing_inode(lc, &ino) == 0 &&
+		  loopcxt_get_backing_devno(lc, &dev) == 0) {
+
+		if (ino == st->st_ino && dev == st->st_dev)
+			goto found;
+
+		/* don't use filename if we have devno and inode */
+		return 0;
+	}
+
+	/* poor man's solution */
+	if (backing_file) {
+		char *name = loopcxt_get_backing_file(lc);
+		int rc = name && strcmp(name, backing_file) == 0;
+
+		free(name);
+		if (rc)
+			goto found;
+	}
+
+	return 0;
+found:
+	if (flags & LOOPDEV_FL_OFFSET) {
+		uint64_t off;
+
+		return loopcxt_get_offset(lc, &off) == 0 && off == offset;
+	}
+	return 1;
+}
+
 /*
  * The setting is removed by loopcxt_set_device() loopcxt_next()!
  */
@@ -630,6 +902,8 @@ int loopcxt_set_offset(struct loopdev_cxt *lc, uint64_t offset)
 	if (!lc)
 		return -EINVAL;
 	lc->info.lo_offset = offset;
+
+	DBG(lc, loopdev_debug("set offset=%jd", offset));
 	return 0;
 }
 
@@ -641,6 +915,8 @@ int loopcxt_set_sizelimit(struct loopdev_cxt *lc, uint64_t sizelimit)
 	if (!lc)
 		return -EINVAL;
 	lc->info.lo_sizelimit = sizelimit;
+
+	DBG(lc, loopdev_debug("set sizelimit=%jd", sizelimit));
 	return 0;
 }
 
@@ -657,6 +933,8 @@ int loopcxt_set_flags(struct loopdev_cxt *lc, uint32_t flags)
 	if (!lc)
 		return -EINVAL;
 	lc->info.lo_flags = flags;
+
+	DBG(lc, loopdev_debug("set flags=%u", (unsigned) flags));
 	return 0;
 }
 
@@ -680,6 +958,7 @@ int loopcxt_set_backing_file(struct loopdev_cxt *lc, const char *filename)
 	strncpy((char *)lc->info.lo_file_name, lc->filename, LO_NAME_SIZE);
 	lc->info.lo_file_name[LO_NAME_SIZE- 1] = '\0';
 
+	DBG(lc, loopdev_debug("set backing file=%s", lc->info.lo_file_name));
 	return 0;
 }
 
@@ -710,6 +989,8 @@ int loopcxt_set_encryption(struct loopdev_cxt *lc,
 	if (!lc)
 		return -EINVAL;
 
+	DBG(lc, loopdev_debug("setting encryption '%s'", encryption));
+
 	if (encryption && *encryption) {
 		if (digits_only(encryption)) {
 			lc->info.lo_encrypt_type = atoi(encryption);
@@ -725,12 +1006,15 @@ int loopcxt_set_encryption(struct loopdev_cxt *lc,
 		lc->info.lo_encrypt_key_size = 0;
 		break;
 	default:
+		DBG(lc, loopdev_debug("setting encryption key"));
 		memset(lc->info.lo_encrypt_key, 0, LO_KEY_SIZE);
 		strncpy((char *)lc->info.lo_encrypt_key, password, LO_KEY_SIZE);
 		lc->info.lo_encrypt_key[LO_KEY_SIZE - 1] = '\0';
 		lc->info.lo_encrypt_key_size = LO_KEY_SIZE;
 		break;
 	}
+
+	DBG(lc, loopdev_debug("encryption successfully set"));
 	return 0;
 }
 
@@ -759,6 +1043,8 @@ int loopcxt_setup_device(struct loopdev_cxt *lc)
 	if (!lc || !*lc->device || !lc->filename)
 		return -EINVAL;
 
+	DBG(lc, loopdev_debug("device setup requested"));
+
 	/*
 	 * Open backing file and device
 	 */
@@ -769,9 +1055,12 @@ int loopcxt_setup_device(struct loopdev_cxt *lc)
 		if (mode != O_RDONLY && (errno == EROFS || errno == EACCES))
 			file_fd = open(lc->filename, mode = O_RDONLY);
 
-		if (file_fd < 0)
+		if (file_fd < 0) {
+			DBG(lc, loopdev_debug("open backing file failed: %m"));
 			return -errno;
+		}
 	}
+	DBG(lc, loopdev_debug("setup: backing file open: OK"));
 
 	if (lc->fd != -1 && lc->mode != mode) {
 		close(lc->fd);
@@ -794,22 +1083,34 @@ int loopcxt_setup_device(struct loopdev_cxt *lc)
 		goto err;
 	}
 
+	DBG(lc, loopdev_debug("setup: device open: OK"));
+
 	/*
 	 * Set FD
 	 */
 	if (ioctl(dev_fd, LOOP_SET_FD, file_fd) < 0) {
 		rc = -errno;
+		DBG(lc, loopdev_debug("LOOP_SET_FD failed: %m"));
 		goto err;
 	}
+
+	DBG(lc, loopdev_debug("setup: LOOP_SET_FD: OK"));
+
 	close(file_fd);
 	file_fd = -1;
 
-	if (ioctl(dev_fd, LOOP_SET_STATUS64, &lc->info))
+	if (ioctl(dev_fd, LOOP_SET_STATUS64, &lc->info)) {
+		DBG(lc, loopdev_debug("LOOP_SET_STATUS64 failed: %m"));
 		goto err;
+	}
+
+	DBG(lc, loopdev_debug("setup: LOOP_SET_STATUS64: OK"));
 
 	memset(&lc->info, 0, sizeof(lc->info));
 	lc->has_info = 0;
+	lc->info_failed = 0;
 
+	DBG(lc, loopdev_debug("setup success [rc=0]"));
 	return 0;
 err:
 	if (file_fd >= 0)
@@ -817,6 +1118,7 @@ err:
 	if (dev_fd >= 0)
 		ioctl(dev_fd, LOOP_CLR_FD, 0);
 
+	DBG(lc, loopdev_debug("setup failed [rc=%d]", rc));
 	return rc;
 }
 
@@ -827,22 +1129,52 @@ int loopcxt_delete_device(struct loopdev_cxt *lc)
 	if (fd < 0)
 		return -EINVAL;
 
-	if (ioctl(fd, LOOP_CLR_FD, 0) < 0)
+	if (ioctl(fd, LOOP_CLR_FD, 0) < 0) {
+		DBG(lc, loopdev_debug("LOOP_CLR_FD failed: %m"));
 		return -errno;
+	}
+
+	DBG(lc, loopdev_debug("device removed"));
 	return 0;
 }
 
+/*
+ * Note that LOOP_CTL_GET_FREE ioctl is supported since kernel 3.1. In older
+ * kernels we have to check all loop devices to found unused one.
+ *
+ * See kernel commit 770fe30a46a12b6fb6b63fbe1737654d28e8484.
+ */
 int loopcxt_find_unused(struct loopdev_cxt *lc)
 {
-	int rc;
+	int rc = -1;
 
-	rc = loopcxt_init_iterator(lc, LOOPITER_FL_FREE);
-	if (rc)
-		return rc;
+	DBG(lc, loopdev_debug("find_unused requested"));
 
-	rc = loopcxt_next(lc);
-	loopcxt_deinit_iterator(lc);
+	if (lc->flags & LOOPDEV_FL_CONTROL) {
+		int ctl = open(_PATH_DEV_LOOPCTL, O_RDWR);
 
+		if (ctl >= 0)
+			rc = ioctl(ctl, LOOP_CTL_GET_FREE);
+		if (rc >= 0) {
+			char name[16];
+			snprintf(name, sizeof(name), "loop%d", rc);
+
+			rc = loopiter_set_device(lc, name);
+		}
+		if (ctl >= 0)
+			close(ctl);
+		DBG(lc, loopdev_debug("find_unused by loop-control [rc=%d]", rc));
+	}
+
+	if (rc < 0) {
+		rc = loopcxt_init_iterator(lc, LOOPITER_FL_FREE);
+		if (rc)
+			return rc;
+
+		rc = loopcxt_next(lc);
+		loopcxt_deinit_iterator(lc);
+		DBG(lc, loopdev_debug("find_unused by scan [rc=%d]", rc));
+	}
 	return rc;
 }
 
@@ -890,32 +1222,19 @@ int loopdev_is_used(const char *device, const char *filename,
 		    uint64_t offset, int flags)
 {
 	struct loopdev_cxt lc;
-	char *backing = NULL;
+	struct stat st;
 	int rc = 0;
 
-	if (!device)
+	if (!device || !filename)
 		return 0;
 
 	loopcxt_init(&lc, 0);
 	loopcxt_set_device(&lc, device);
 
-	backing = loopcxt_get_backing_file(&lc);
-	if (!backing)
-		goto done;
-	if (filename && strcmp(filename, backing) != 0)
-		goto done;
-	if (flags & LOOPDEV_FL_OFFSET) {
-		uint64_t off;
+	rc = !stat(filename, &st);
+	rc = loopcxt_is_used(&lc, rc ? &st : NULL, filename, offset, flags);
 
-		if (loopcxt_get_offset(&lc, &off) != 0 || off != offset)
-			goto done;
-	}
-
-	rc = 1;
-done:
-	free(backing);
 	loopcxt_deinit(&lc);
-
 	return rc;
 }
 
@@ -924,6 +1243,9 @@ int loopdev_delete(const char *device)
 	struct loopdev_cxt lc;
 	int rc;
 
+	if (!device)
+		return -EINVAL;
+
 	loopcxt_init(&lc, 0);
 	rc = loopcxt_set_device(&lc, device);
 	if (!rc)
@@ -938,33 +1260,23 @@ int loopdev_delete(const char *device)
 int loopcxt_find_by_backing_file(struct loopdev_cxt *lc, const char *filename,
 				 uint64_t offset, int flags)
 {
-	int rc;
+	int rc, hasst;
+	struct stat st;
 
 	if (!filename)
 		return -EINVAL;
 
+	hasst = !stat(filename, &st);
+
 	rc = loopcxt_init_iterator(lc, LOOPITER_FL_USED);
 	if (rc)
 		return rc;
 
-	while((rc = loopcxt_next(lc)) == 0) {
-		char *backing = loopcxt_get_backing_file(lc);
+	while ((rc = loopcxt_next(lc)) == 0) {
 
-		if (!backing || strcmp(backing, filename)) {
-			free(backing);
-			continue;
-		}
-
-		free(backing);
-
-		if (flags & LOOPDEV_FL_OFFSET) {
-			uint64_t off;
-			if (loopcxt_get_offset(lc, &off) != 0 || offset != off)
-				continue;
-		}
-
-		rc = 0;
-		break;
+		if (loopcxt_is_used(lc, hasst ? &st : NULL,
+					filename, offset, flags))
+			break;
 	}
 
 	loopcxt_deinit_iterator(lc);
@@ -990,18 +1302,60 @@ char *loopdev_find_by_backing_file(const char *filename, uint64_t offset, int fl
 	return res;
 }
 
+/*
+ * Returns number of loop devices associated with @file, if only one loop
+ * device is associeted with the given @filename and @loopdev is not NULL then
+ * @loopdev returns name of the device.
+ */
+int loopdev_count_by_backing_file(const char *filename, char **loopdev)
+{
+	struct loopdev_cxt lc;
+	int count = 0;
+
+	if (!filename)
+		return -1;
+
+	loopcxt_init(&lc, 0);
+	if (loopcxt_init_iterator(&lc, LOOPITER_FL_USED))
+		return -1;
+
+	while(loopcxt_next(&lc) == 0) {
+		char *backing = loopcxt_get_backing_file(&lc);
+
+		if (!backing || strcmp(backing, filename)) {
+			free(backing);
+			continue;
+		}
+
+		free(backing);
+		if (loopdev && count == 0)
+			*loopdev = loopcxt_strdup_device(&lc);
+		count++;
+	}
+
+	loopcxt_deinit(&lc);
+
+	if (loopdev && count > 1) {
+		free(*loopdev);
+		*loopdev = NULL;
+	}
+	return count;
+}
+
 
 #ifdef TEST_PROGRAM_LOOPDEV
 #include <errno.h>
 #include <err.h>
 
-static void test_loop_info(const char *device, int flags)
+static void test_loop_info(const char *device, int flags, int debug)
 {
 	struct loopdev_cxt lc;
 	char *p;
 	uint64_t u64;
 
 	loopcxt_init(&lc, flags);
+	loopcxt_enable_debug(&lc, debug);
+
 	if (loopcxt_set_device(&lc, device))
 		err(EXIT_FAILURE, "failed to set device");
 
@@ -1020,12 +1374,13 @@ static void test_loop_info(const char *device, int flags)
 	loopcxt_deinit(&lc);
 }
 
-static void test_loop_scan(int flags)
+static void test_loop_scan(int flags, int debug)
 {
 	struct loopdev_cxt lc;
 	int rc;
 
 	loopcxt_init(&lc, 0);
+	loopcxt_enable_debug(&lc, debug);
 
 	if (loopcxt_init_iterator(&lc, flags))
 		err(EXIT_FAILURE, "iterator initlization failed");
@@ -1047,12 +1402,13 @@ static void test_loop_scan(int flags)
 	loopcxt_deinit(&lc);
 }
 
-static int test_loop_setup(const char *filename, const char *device)
+static int test_loop_setup(const char *filename, const char *device, int debug)
 {
 	struct loopdev_cxt lc;
 	int rc = 0;
 
 	loopcxt_init(&lc, 0);
+	loopcxt_enable_debug(&lc, debug);
 
 	if (device) {
 		rc = loopcxt_set_device(&lc, device);
@@ -1089,27 +1445,31 @@ static int test_loop_setup(const char *filename, const char *device)
 
 int main(int argc, char *argv[])
 {
+	int dbg;
+
 	if (argc < 2)
 		goto usage;
 
+	dbg = getenv("LOOPDEV_DEBUG") == NULL ? 0 : 1;
+
 	if (argc == 3 && strcmp(argv[1], "--info") == 0) {
 		printf("---sysfs & ioctl:---\n");
-		test_loop_info(argv[2], 0);
+		test_loop_info(argv[2], 0, dbg);
 		printf("---sysfs only:---\n");
-		test_loop_info(argv[2], LOOPDEV_FL_NOIOCTL);
+		test_loop_info(argv[2], LOOPDEV_FL_NOIOCTL, dbg);
 		printf("---ioctl only:---\n");
-		test_loop_info(argv[2], LOOPDEV_FL_NOSYSFS);
+		test_loop_info(argv[2], LOOPDEV_FL_NOSYSFS, dbg);
 
 	} else if (argc == 2 && strcmp(argv[1], "--used") == 0) {
 		printf("---all used devices---\n");
-		test_loop_scan(LOOPITER_FL_USED);
+		test_loop_scan(LOOPITER_FL_USED, dbg);
 
 	} else if (argc == 2 && strcmp(argv[1], "--free") == 0) {
 		printf("---all free devices---\n");
-		test_loop_scan(LOOPITER_FL_FREE);
+		test_loop_scan(LOOPITER_FL_FREE, dbg);
 
 	} else if (argc >= 3 && strcmp(argv[1], "--setup") == 0) {
-		test_loop_setup(argv[2], argv[3]);
+		test_loop_setup(argv[2], argv[3], dbg);
 
 	} else if (argc == 3 && strcmp(argv[1], "--delete") == 0) {
 		if (loopdev_delete(argv[2]))
